#!/usr/bin/python2
# port of ppcasm's N64 HomeBrew Loader

from serial import Serial
import sys
from struct import pack
from binascii import a2b_hex

def sendIntro(quiet=False):
  port.write("\xde\xad\xbe\xef")
  if not quiet:
    print '> ' + port.readline(),

def readRAM(size, address, quiet=False):
  port.write(pack('>LL', size|0x80000000, address))

  while True:
    line = port.readline()
    if line[0] == '$':
      if not quiet:
        print '$%08x='%address + line,
      break
    else:
      if not quiet:
        print '> ' + line,

  return a2b_hex(line[1:].rstrip())

def writeRAM(address, payload, quiet=False):
  port.write(pack('>LL', len(payload), address) + payload)

def sendOutro():
  port.write(pack('>L', 0))

def dumpPort(quiet=False):
  while port.inWaiting() > 0:
    print port.readline()

# mips.h stuff for generating code

MIPS_R0 = 0
MIPS_AT = 1
MIPS_V0 = 2
MIPS_V1 = 3
MIPS_A0 = 4
MIPS_A1 = 5
MIPS_A2 = 6
MIPS_A3 = 7
MIPS_T0 = 8
MIPS_T1 = 9
MIPS_T2 = 10
MIPS_T3 = 11
MIPS_T4 = 12
MIPS_T5 = 13
MIPS_T6 = 14
MIPS_T7 = 15
MIPS_S0 = 16
MIPS_S1 = 17
MIPS_S2 = 18
MIPS_S3 = 19
MIPS_S4 = 20
MIPS_S5 = 21
MIPS_S6 = 22
MIPS_S7 = 23
MIPS_T8 = 24
MIPS_T9 = 25
MIPS_K0 = 26
MIPS_K1 = 27
MIPS_GP = 28
MIPS_SP = 29
MIPS_FP = 30
MIPS_RA = 31


def OP(opcode, special=0):
  return (opcode << 26) | special
def RD(x):
  return x << 11
def RT(x):
  return x << 16
def RS(x):
  return x << 21
def IM(x):
  return x&0xffff
def JT(x):
  return (x>>2)&0x3ffffff

def MFC0(rt, rd): return OP(0x10) | RS(0x00) | RT(rt) | RD(rd)
def MTC0(rt, rd): return OP(0x10) | RS(0x04) | RT(rt) | RD(rd)
def ADDIU(rt, rs, immd): return OP(0x09) | RS(rs) | RT(rt) | IM(immd)
def ORI(rt, rs, immd):   return OP(0x0d) | RS(rs) | RT(rt) | IM(immd)
def AND(rd, rs, rt):     return OP(0x00, 0x24) | RS(rs) | RT(rt) | RD(rd)
def LUI(rt, immd):       return OP(0x0F) | RT(rt) | IM(immd)
def SW(rt, off, base):   return OP(0x2B) | RS(base) | RT(rt) | IM(off)
ERET = OP(0x10, 0x18) | (1 << 25)
SYNC = 0xf
NOP = 0


def makejump(address): return OP(2) | JT(address)

INSN_PATCH_ADDR = 0xA07C5C00
GLOBAL_OFFSET_TABLE = 0xA0000200


# build loader

UPLOAD_ADDR = 0x80000400
EMBED_ADDR = 0x80300000

loader = ''.join(pack(">L", dw) for dw in \
[

  # Stop GameShark traps 
  MTC0(MIPS_R0, 18),
  NOP,
  MTC0(MIPS_R0, 19),
  NOP,
    
  # Disable Interrupts
  MFC0(MIPS_T0, 12),
  ADDIU(MIPS_T1, MIPS_R0, 0xfffe),
  AND(MIPS_T0, MIPS_T0, MIPS_T1),
  MTC0(MIPS_T0, 12),
  
  # Modify EPC
  LUI(MIPS_K0, UPLOAD_ADDR>>16),
  ORI(MIPS_K0, MIPS_K0, UPLOAD_ADDR),
  NOP,
  MTC0(MIPS_K0, 14),
  NOP,
     
  # Patch back modified code handler
  LUI(MIPS_K1, 0x3c1a),
  ORI(MIPS_K1, MIPS_K1, 0x8000),
  NOP,
  LUI(MIPS_K0, 0xa07c),
  ORI(MIPS_K0, MIPS_K0, 0x5c00),
  NOP,
  SW(MIPS_K1, 0, MIPS_K0),
  NOP,
  SYNC,
  NOP,

  # Halt RSP
  LUI(MIPS_T1, 2),
  LUI(MIPS_T0, 0xa404),
  ORI(MIPS_T0, MIPS_T0, 0x0010),
  NOP,
  SW(MIPS_T1, 0, MIPS_T0),
  NOP,

  # Halt RDP
  LUI(MIPS_T1, 1|4|0x10|0x40|0x80|0x100|0x200),
  LUI(MIPS_T0, 0xa410),
  ORI(MIPS_T0, MIPS_T0, 0x000c),
  NOP,
  SW(MIPS_T1, 0, MIPS_T0),
  NOP,

  # Return from interrupt - execute code
  ERET,
  NOP
])

def hex_string(s):
  if type(s) == int:
    return '%02x'%s
  else:
    return ''.join('%02x'%ord(c) for c in s)


######


#s = hex_string(loader)
#for i in range(0,len(s),8):
#  print s[i:i+8]

deviceName = "/dev/ttyACM1"
port = Serial(deviceName, 57600)

infile = open(sys.argv[1], "rb")
data = infile.read()
infile.close()

sendIntro()
dumpPort()

# upload payload
print "send payload"
writeRAM(UPLOAD_ADDR, data)
dumpPort()
print "done"

sendOutro()
port.close()


###
port.open()
sendIntro()

# upload embedded code
print "send embedded code"
writeRAM(EMBED_ADDR, loader)
dumpPort()
print "done"

sendOutro()
port.close()

port.open()
sendIntro()

# set patch
print "set patch"
writeRAM(INSN_PATCH_ADDR, pack(">L", makejump(EMBED_ADDR)))
print "done"
dumpPort()

# not sure why this is necessary
readRAM(4, 0xA0000000)

sendOutro()
dumpPort()

dumpPort()
port.close()


